/* Pxys - An IRC ProxyScan, ms_network.c
 * 
 * Copyright (C) 2001 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 *
 * Net junction/split handlers.
 *
 * $Id: ms_network.c,v 1.12 2001/10/31 20:49:19 mbuna Exp $
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "m_array.h"
#include "m_share.h"
#include "m_structs.h"
#include "m_noscan.h"

#include "ms_network.h"
#include "ms_netutils.h"
#include "ms_globals.h"
#include "ms_msgout.h"

#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "misc.h"
#include "numnicks.h"
#include "strnlwrcpy.h"
#include "typedefs.h"
#include "xmalloc.h"
#include "yxx.h"

#include "m_log.h"

struct Server *mServer = NULL;
extern pthread_mutex_t mutex_server;

long serv_current = 0;

htab *mHS;
server_handle_t mSA;

void
network_init()
{
   /* Create main Hashed Servers table */
   mHS = jenkins_hcreate(2);
   
   /* Pre-allocate FULL servers array. */
   mSA = (server_handle_t) xcalloc(NN_MAX_SERVER, sizeof(void *));
   
   serv_current = 0;
}

void
server_add(up_nserv, name, nserv, ncap)
unsigned int up_nserv;   /* Uplink numeric */
char *name;              /* Server name */
unsigned int nserv;      /* Server numeric ID */
unsigned int ncap;       /* Server capacity (max clients) */
{
   struct Server *newServer;
   unsigned int max_clients = 16;
   
   while (max_clients < ncap)
      max_clients <<= 1;
   --max_clients;
   
   newServer = (struct Server *) xmalloc(sizeof(struct Server));

   /* Server name (we use lowercases in hash table) */
   newServer->name = (char *) xmalloc(strlen(name) + 1);
   strnlwrcpy(newServer->name, name, strlen(name) + 1);
   
   /* Uplink numeric */
   newServer->upx = up_nserv;
   
   /* Server numeric */
   newServer->myx = nserv;
   newServer->nn_mask = max_clients;
   
   newServer->noscan = false;
   newServer->shared = false;
   
   /* Allocate clients numerics array */
   newServer->cliH = (void**) xcalloc(max_clients + 1, sizeof(void*));
   
   pthread_mutex_lock(&mutex_server);
   
   mSA[nserv] = newServer;
   ++serv_current;
   
   hadd(mHS, newServer->name, strlen(newServer->name) + 1, (void*) newServer);
    
   /* Refresh noscan and shared servers list. */
   refresh_noscan_servers();
   refresh_shared_servers();
   
   pthread_mutex_unlock(&mutex_server);
   
   Debug((DBLBASIC, "server_add: server %s added (capacity: %u)",
        newServer->name, max_clients + 1));
   
   if(chk_shared(name)) {
      sendto_console("[%s] Sharing with %s activated.", Me->id, name) ;
   }
   /*
   if (MyGline->direct)
      gonnaSpreadGlines = 2;
      */
}

/* ------------------- NET SPLITS HANDLING ------------------- */


void
do_server_quit(char *name)
{
   if(hfind(mHS, strlwr(name), strlen(name) + 1) == TRUE) {
      
      Debug((DBLMAX, "do_server_quit: server %s found.",
         ((struct Server *)hstuff(mHS))->name));
         
      do_squit((struct Server *)hstuff(mHS));
      
   } else {
      
      Debug((DBLMAX, "current pos: %s",
         ((struct Server *)hstuff(mHS))->name));
      
      Debug((DBLMAX, "do_server_quit: server %s not found.", name));
      log_error("do_server_quit: server %s not found.\n", name);
   
   }
}

void
do_squit(sptr)
struct Server *sptr;
{
   pthread_mutex_lock(&mutex_server);
   
   do_recursive_squit(sptr);
   
   refresh_shared_servers();   /* Update sharing toggles */
   refresh_relay_services();
   
   pthread_mutex_unlock(&mutex_server);
}

void
do_recursive_squit(sptr)
struct Server *sptr;
{
   int rs = 0;        /* rescan server hash table flag */
   
   Debug((DBLMAX, "do_recursive_squit> %s", sptr->name));
   
   hdel(mHS);
   
   if(chk_shared(sptr->name))
      sendto_console("[%s] Sharing with %s disabled.", Me->id, sptr->name) ;
   
   /* Delete all the users from this server. */
   array_purgeserv(sptr->myx);
   
   /* Process recursively to purge downlinks. */
   if (hfirst(mHS)) do {
      rs = 0;
      if(((struct Server *)hstuff(mHS))->upx == sptr->myx) {
         do_recursive_squit((struct Server *)hstuff(mHS));
         /* restart loop -- cheap way to avoid collision or double delete */
         if(!hfirst(mHS))
            break;
         rs = 1; /* reloop */
      }
   } while (rs || hnext(mHS)); /* reloop from start or try next */
   
   /* Really delete now this server. */
   mSA[sptr->myx] = NULL;
   --serv_current;
   
   xfree(sptr->name);
   xfree(sptr->cliH);
   xfree(sptr);
   
   Debug((DBLBASIC, "do_recursive_squit> Server data (%u) free'd.", sptr->myx));
}

void
refresh_relay_services()
{
   struct Relay_s *relay;
   
   for (relay = MyGline->relay; relay; relay = relay->next) {
      if(!hfind(mHS, relay->serv, strlen(relay->serv) + 1)) {
         /* relay cannot be found -- set it as "deauthed" */
         relay->isauth = 0;
      }
   }
}


/* ms_network.c */
