/* Pxys - An IRC ProxyScan, ms_socket.c
 * 
 * Copyright (C) 2001 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 *
 * IRC server<>server connection.
 *
 * $Id: ms_socket.c,v 1.6 2001/10/29 19:24:00 mbuna Exp $
 */
 
#define IRC_THREAD_STACKSIZE 262144
#include "config.h"

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/resource.h>  
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#include <errno.h>
#include <string.h>

#include "ms_auth.h"
#include "ms_globals.h"
#include "ms_msgout.h"
#include "ms_parser.h"
#include "ms_socket.h"

#include "mp_utils.h"           /* set_sockbufs, set_nonblock_flag */

#include "m_thread.h"
#include "m_timers.h"

#include "typedefs.h"
#include "misc.h"

#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "xmalloc.h"


pthread_mutex_t w_mutex;

extern htab *mHC;

struct Uplink_s *curUplink = NULL;
int filedes;
time_t connection_starttime = 0;


#define BUFSIZE         512     /* RFC 1459 */
#define IsEol(c) (c == '\r') || (c == '\n')

static char linebuf[BUFSIZE];
long recvBytes = 0;
size_t msgoutSize = 0;

#ifdef RLIMIT_FDMAX
#define RLIMIT_FD_MAX   RLIMIT_FDMAX
#else
#ifdef RLIMIT_NOFILE
#define RLIMIT_FD_MAX RLIMIT_NOFILE
#else
#ifdef RLIMIT_OPEN_MAX
#define RLIMIT_FD_MAX RLIMIT_OPEN_MAX
#else
#undef RLIMIT_FD_MAX
#endif
#endif
#endif

void
irc_start()
{
   pthread_t ircthread;
   struct rlimit limit;
   pthread_attr_t attr;
   size_t size;
   
   pthread_attr_init(&attr);
   pthread_attr_getstacksize(&attr, (size_t*)&size);
   Debug((DBLMAX, "stacksize: %d\n", size));
   if(size < IRC_THREAD_STACKSIZE)
      pthread_attr_setstacksize(&attr, IRC_THREAD_STACKSIZE);
   
   pthread_attr_getstacksize(&attr, (size_t*)&size);
   
   Debug((DBLMAX, "new stacksize: %d\n", size));
   
#ifdef RLIMIT_SBSIZE
   if (!getrlimit(RLIMIT_SBSIZE, &limit)) {
      limit.rlim_cur = limit.rlim_max;    /* make soft limit the max */
      setrlimit(RLIMIT_SBSIZE, &limit);
   }
#endif
   
#ifdef RLIMIT_FD_MAX
   if (!getrlimit(RLIMIT_FD_MAX, &limit)) {
      limit.rlim_cur = limit.rlim_max;    /* make soft limit the max */
      setrlimit(RLIMIT_FD_MAX, &limit);
   }
#endif
   
   if(pthread_create(&ircthread, &attr, RunIRC, NULL)) {
      fprintf(stderr, "IRC Server thread failed. Cannot continue.\n\n");
      exit(EXIT_FAILURE);
   }
   printf("irc_start: IRC Server thread (%x) created and started.\n", (unsigned int)ircthread);
   
   pthread_join(ircthread, NULL);
}




void
irc_init()
{
   struct Uplink_s *searchUp;
   
   /* Choose uplink server... */
   if(NULL == curUplink) {
      /* Always the first case: choose the uplink with priority of 1 (highest) */
      for(searchUp = MyUplink; searchUp; searchUp = searchUp->next) {
         if(searchUp->priority <= 1) {
            curUplink = searchUp;
            break;
         }
      }
   } else {
      /* 1 is high priority, more is low priority */
      for(searchUp = MyUplink; searchUp; searchUp = searchUp->next) {
         if(searchUp->priority > curUplink->priority) {
            if(searchUp != curUplink) {
               curUplink = searchUp;
               break;
            }
         }
      }
      if(NULL == searchUp) {
       /* Not found, come back to high priority uplink */
         for(searchUp = MyUplink; searchUp; searchUp = searchUp->next) {
            if(searchUp->priority <= 1) {
               curUplink = searchUp;
               break;
            }
         }
      }
   }
}

void
irc_setsocket(struct sockaddr_in *servAddrPtr)
{
   struct sockaddr_in localAddr;
   struct hostent *h;
   struct linger setLinger;
   unsigned int optval;
   int res;
   char **p;
   
   h = gethostbyname(curUplink->host);

   if(h == NULL) {
      fprintf(stderr, "pxys: unknown host '%s'\n", curUplink->host);
      exit(1);
   }
   
   /* create socket */
   filedes = socket(AF_INET, SOCK_STREAM, 0);
   if(filedes < 0) {
      fprintf(stderr, "pxys: ms_socket.c: cannot open socket, aborting.\n");
      exit(1);
   }
   
   memset(servAddrPtr, 0, sizeof(struct sockaddr_in));
   servAddrPtr->sin_family = AF_INET; /* h->h_addrtype; */
   p = h->h_addr_list ;
   memcpy((void *) &servAddrPtr->sin_addr.s_addr, *p, h->h_length);
   servAddrPtr->sin_port = htons(curUplink->port);
   
   /* bind only if we are using a virtual host */
   if(curUplink->vhost) {
      localAddr.sin_family = AF_INET;
      localAddr.sin_addr.s_addr = curUplink->bindaddr;
      localAddr.sin_port = htons(0);
   }
   
   setLinger.l_onoff = 0;
   setLinger.l_linger = 0;
   
   setsockopt(filedes, SOL_SOCKET, SO_LINGER, (const char*)&setLinger, sizeof(struct linger));
   
   optval = 1;
   setsockopt(filedes, SOL_SOCKET, SO_KEEPALIVE, (const char*)&optval, sizeof(int));
   
   optval = 1;
   setsockopt(filedes, SOL_SOCKET, SO_REUSEADDR, (const char*)&optval, sizeof(optval));
   /*
   optval = 1;
   if (setsockopt(filedes, IPPROTO_TCP, TCP_NODELAY, (void *) &optval, sizeof(optval))) {
   */
   if(!set_sockbufs(filedes, 64 * 1024)) {   /* 64kb */
      fprintf(stderr, "irc_start: Unable to set socket buffers size.\n");
      exit(1);
   }
   
   if(set_nonblock_flag(filedes) == -1) {
      fprintf(stderr, "irc_start: Unable to set socket NONBLOCK.\n");
      exit(1);
   }
   
   
   if(curUplink->vhost) {
      res = bind(filedes, (struct sockaddr *) &localAddr, sizeof(localAddr));
      if(res < 0) {
         fprintf(stderr, "irc_start: Cannot bind on IP (vhost): %s\n", curUplink->bindip);
         exit(1);
      }
#ifdef DEBUG
      else
         Debug((DBLMAX, "irc_start: Bind to %s\n", curUplink->bindip));
#endif
   }

   printf("Connecting to %s.\n", inet_ntoa(servAddrPtr->sin_addr));
}


void*
RunIRC()
{
   int bufSize;
   int length = 0;
   char *readbuf;
   struct sockaddr_in servAddr;
   
   
   pthread_mutex_init(&w_mutex, NULL);
   
   irc_init();
   
   irc_setsocket(&servAddr);
   
   /* connect to server */
   if(!connect_nonb(filedes, &servAddr)) {
      fprintf(stderr, "irc_start: Cannot connect to %s:%u, aborting.\n",
      	curUplink->host, curUplink->port);
      myclose(filedes);
      pthread_mutex_destroy(&w_mutex);
      return NULL;
   }
   
   if(wait_for_connection(filedes) < 0) {
      fprintf(stderr, "Cannot connect to %s:%u, aborting\n", curUplink->host, curUplink->port);
      myclose(filedes);
      pthread_mutex_destroy(&w_mutex);
      return NULL;
   }
   
   time(&connection_starttime);
   
   bufSize = get_recv_bufSize(filedes);
   printf("Connected. TCP window size: %d bytes.\n", bufSize);
   
   // Creating an input buffer == size of the socket input buffer
   readbuf = (char *) xmalloc(bufSize * sizeof(char));
   
   
   msgoutSize = 0;
   
   
   /* Send IRC authentication */
   irc_auth(curUplink);
   
   
   /* Grab data from server */
   do {
      
      length = sock_select(filedes, readbuf, bufSize);
      
      if(length > 0) {
         recvBytes += length;
         depacket(readbuf, length);
      }
       
   } while (length >= 0);
   
   /* sock_select failed... */
   
   myclose(filedes);
   
   Debug((DBLMAX, "Connection with %s:%u closed.\n", curUplink->host, curUplink->port));
   
   xfree((void*)readbuf);
   
   pthread_mutex_destroy(&w_mutex);
   return NULL;
}

int
sock_select(int fd, char *readbuf, size_t bufSize)
{
   short int cnt = 10;
   int res = -1;
   fd_set readfds, writefds;
   fd_set* wfdsp = NULL;     /* default: nothing to write */
   struct timeval tv;
   
   FD_ZERO(&readfds);
   FD_SET(fd, &readfds) ;
   
   if (msgoutSize > 0) {
      wfdsp = &writefds;
      FD_ZERO(wfdsp);
      FD_SET(fd, wfdsp);
   }
   
   do {
      errno = 0 ;
      tv.tv_sec	= 1; /* select may change tv on linux */
      tv.tv_usec = 0;
      res = select(fd + 1, &readfds, wfdsp, NULL, &tv) ;
   } while((errno == EINTR) && (--cnt >= 0)) ;
   
   if(res < 0) {
      /* select error */
      Debug((DBLMAX, "sock_select: select error\n"));
      return res;
   }
   
   m_timers_check();
   
   if(res == 0)
      /* Timeout */
      return res;
   
   if(wfdsp && FD_ISSET(fd, wfdsp)) {
      errno = 0;
      res = send_queued_now(fd);
      if(res < 0)
         return res;
      /*fall thru*/
   }
   
   if(FD_ISSET(fd, &readfds)) {
      do {
         errno = 0 ;
         res = read(fd, readbuf, bufSize);
         
         if(res <= 0) {
            if (EWOULDBLOCK == errno || EAGAIN == errno)
               return 0;
            return -1;
         }
      } while((--cnt > 0) && (EINTR == errno)) ;
   }
   
   return res;
}


void
depacket(const char *buffer, int length)
{
   static unsigned int bufpos = 0;
   const char* src = buffer;
   char *p;
   
   p = linebuf + bufpos;
   
   while (length-- > 0) {
      *p = *src++;
      
      if(IsEol(*p)) {
         if(p == linebuf)
            continue;
         *p = '\0';
         
         irc_parse(linebuf);
         p = linebuf;
      
      } else if(p < linebuf + BUFSIZE)
         ++p;
   }
   bufpos = p - linebuf;
}

int
get_recv_bufSize(int fd)
{
   int recvSize = 65536 ;

   /* socklen_t */
   unsigned int len = sizeof(recvSize);
   
   if(getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (void *) &recvSize, &len) < 0) {
      return -1;
   }
   
   return recvSize;
}

int
wait_for_connection(int fd)
{
   struct timeval tv;
   int fdcnt = -1, cnt=10;
   
   fd_set writefds;
   
   FD_ZERO( &writefds );
   FD_SET( fd, &writefds );
   
   tv.tv_sec = 10;
   tv.tv_usec = 0;
   
   do {
      errno = 0 ;
      fdcnt = select(fd + 1, NULL, &writefds, NULL, &tv) ;
      tv.tv_sec = 10;
      tv.tv_usec = 0;
   } while( (errno == EINTR) && (--cnt >= 0) ) ;
   
   if(fdcnt < 0)
      return fdcnt;
   
   return get_sockerr(fd);
}



/* ms_socket.c */
